==> clickhouse-0.12.1.tar: lib/clickhouse.ex <==
defmodule ClickHouse do
  @moduledoc """
  A ClickHouse client.

  This currently represents an early work in progress.
  """

  use Supervisor

  alias ClickHouse.{Config, Query, Result}

  @opts_schema %{
    name: [type: :atom, required: true, default: :default],
    interface: [type: :module, required: true, default: ClickHouse.Interface.HTTP],
    formats: [
      type: {:list, :atom},
      required: true,
      default: [
        ClickHouse.Format.TSV,
        ClickHouse.Format.TSVWithNames,
        ClickHouse.Format.TSVWithNamesAndTypes,
        ClickHouse.Format.Values
      ]
    ]
  }

  @typedoc """
  The client identifier.
  """
  @type client :: atom()

  @typedoc """
  A query statement.
  """
  @type statement :: binary()

  @typedoc """
  Parameters used with queries.
  """
  @type params :: list()

  @typedoc """
  Options used for `child_spec/1` and `start_link/1`
  """
  @type option ::
          {:name, client()}
          | {:interface, ClickHouse.Interface.t()}
          | {:formats, [ClickHouse.Format.t()]}

  @typedoc """
  Various representations of ClickHouse-related errors.
  """
  @type error ::
          ClickHouse.ConnectionError.t()
          | ClickHouse.CoordinationError.t()
          | ClickHouse.DatabaseError.t()
          | ClickHouse.NetworkError.t()
          | ClickHouse.ParsingError.t()

  ################################
  # Public API
  ################################

  @doc """
  Starts a ClickHouse client.

  ## Options

    * `:name` - The name of the client - defaults to `:default`.
    * `:interface` - The network interface to use for queries - defaults to `ClickHouse.Interface.HTTP`.
    * `:formats` - A list of formats available for encoding/decoding.

  ## Extra Options

  Any additional options passed will be given to the client interface.
  """
  @spec start_link([option()]) :: Supervisor.on_start()
  def start_link(opts \\ []) do
    {opts, extra_opts} = Keyword.split(opts, Map.keys(@opts_schema))
    opts = KeywordValidator.validate!(opts, @opts_schema)
    Supervisor.start_link(__MODULE__, {opts, extra_opts})
  end

  @doc """
  Prepares and executes a query using a ClickHouse client.
  """
  @spec query(client(), statement(), params(), keyword()) ::
          {:ok, ClickHouse.Result.t()} | {:error, error()}
  def query(client, statement, params \\ [], opts \\ []) do
    query = prepare(client, statement, params, opts)
    execute(query)
  end

  @doc """
  Prepares a query for a ClickHouse client.
  """
  @spec prepare(client(), statement(), params(), keyword()) :: ClickHouse.Query.t()
  def prepare(client, statement, params \\ [], opts \\ []) do
    client
    |> Query.new(statement, params, opts)
    |> Query.encode()
  end

  @doc """
  Executes a query using a ClickHouse client.
  """
  @spec execute(ClickHouse.Query.t()) :: {:ok, ClickHouse.Result.t()} | {:error, error()}
  def execute(%{interface: nil}) do
    raise(ArgumentError, "No network interface specified")
  end

  def execute(query) do
    with {:ok, result} <- query.interface.execute(query) do
      result = Result.decode(result)
      {:ok, result}
    end
  end

  ################################
  # Supervisor Callbacks
  ################################

  @impl Supervisor
  def init({opts, extra_opts}) do
    init_config(opts)

    children = [
      interface(opts, extra_opts)
    ]

    Supervisor.init(children, strategy: :one_for_one)
  end

  ################################
  # Private API
  ################################

  defp init_config(opts) do
    client = Keyword.fetch!(opts, :name)
    Config.new(client, opts)
  end

  defp interface(opts, extra_opts) do
    client = Keyword.fetch!(opts, :name)
    interface = Keyword.fetch!(opts, :interface)
    interface_opts = Keyword.put(extra_opts, :name, client)

    {interface, interface_opts}
  end
end

==> clickhouse-0.13.1.tar: lib/clickhouse.ex <==
defmodule ClickHouse do
  @moduledoc """
  A ClickHouse client.

  This currently represents an early work in progress.
  """

  use Supervisor

  alias ClickHouse.{Client, Query, Result}

  @opts_schema %{
    name: [type: :atom, required: true, default: :default],
    interface: [type: :module, required: true, default: ClickHouse.Interface.HTTP],
    formats: [
      type: {:list, :atom},
      required: true,
      default: [
        ClickHouse.Format.TSV,
        ClickHouse.Format.TSVWithNames,
        ClickHouse.Format.TSVWithNamesAndTypes,
        ClickHouse.Format.Values
      ]
    ]
  }

  @typedoc """
  A ClickHouse client.
  """
  @type client :: atom() | ClickHouse.Client.t()

  @typedoc """
  A query statement.
  """
  @type statement :: binary()

  @typedoc """
  Parameters used with queries.
  """
  @type params :: list()

  @typedoc """
  Options used for `child_spec/1` and `start_link/1`
  """
  @type option ::
          {:name, atom()}
          | {:interface, ClickHouse.Interface.t()}
          | {:formats, [ClickHouse.Format.t()]}

  @typedoc """
  Various representations of ClickHouse-related errors.
  """
  @type error ::
          ClickHouse.ConnectionError.t()
          | ClickHouse.CoordinationError.t()
          | ClickHouse.DatabaseError.t()
          | ClickHouse.NetworkError.t()
          | ClickHouse.ParsingError.t()

  ################################
  # Public API
  ################################

  @doc """
  Starts a ClickHouse client.

  ## Options

    * `:name` - The name of the client - defaults to `:default`.
    * `:interface` - The network interface to use for queries - defaults to `ClickHouse.Interface.HTTP`.
    * `:formats` - A list of formats available for encoding/decoding.

  ## Extra Options

  Any additional options passed will be given to the client interface.
  """
  @spec start_link([option()]) :: Supervisor.on_start()
  def start_link(opts \\ []) do
    {opts, extra_opts} = Keyword.split(opts, Map.keys(@opts_schema))
    opts = KeywordValidator.validate!(opts, @opts_schema)
    Supervisor.start_link(__MODULE__, {opts, extra_opts})
  end

  @doc """
  Prepares and executes a query using a ClickHouse client.
  """
  @spec query(client(), statement(), params(), keyword()) ::
          {:ok, ClickHouse.Result.t()} | {:error, error()}
  def query(client, statement, params \\ [], opts \\ [])

  def query(%Client{} = client, statement, params, opts) do
    query = prepare(client, statement, params)
    execute(client, query, opts)
  end

  def query(client, statement, params, opts) when is_atom(client) do
    client = fetch_client!(client)
    query(client, statement, params, opts)
  end

  @doc """
  Prepares a query for a ClickHouse client.
  """
  @spec prepare(client(), statement(), params()) :: ClickHouse.Query.t()
  def prepare(client, statement, params \\ [])

  def prepare(%Client{} = client, statement, params) do
    Query.prepare(client, statement, params)
  end

  def prepare(client, statement, params) when is_atom(client) do
    client = fetch_client!(client)
    prepare(client, statement, params)
  end

  @doc """
  Executes a query using a ClickHouse client.
  """
  @spec execute(client(), ClickHouse.Query.t(), opts :: keyword()) ::
          {:ok, ClickHouse.Result.t()} | {:error, error()}
  def execute(client, query, opts \\ [])

  def execute(%Client{} = client, query, opts) do
    with {:ok, result} <- client.interface.execute(client, query, opts) do
      result = Result.decode(result)
      {:ok, result}
    end
  end

  def execute(client, query, opts) when is_atom(client) do
    client = fetch_client!(client)
    execute(client, query, opts)
  end

  ################################
  # Supervisor Callbacks
  ################################

  @impl Supervisor
  def init({opts, extra_opts}) do
    init_client(opts)

    children = [
      interface(opts, extra_opts)
    ]

    Supervisor.init(children, strategy: :one_for_one)
  end

  ################################
  # Private API
  ################################

  defp init_client(opts) do
    client = Client.new(opts)
    :ok = :persistent_term.put({__MODULE__, client.name}, client)
  end

  defp fetch_client!(name) do
    :persistent_term.get({__MODULE__, name})
  rescue
    ArgumentError ->
      raise ArgumentError, """
      No ClickHouse client #{inspect(name)} available.
      """
  end

  defp interface(opts, extra_opts) do
    client = Keyword.fetch!(opts, :name)
    interface = Keyword.fetch!(opts, :interface)
    interface_opts = Keyword.put(extra_opts, :name, client)

    {interface, interface_opts}
  end
end
